import 'package:meta/meta.dart';
import 'package:floop/internals.dart';

import './flutter_import.dart';

export 'package:floop/floop.dart' show ObservedMap;

var store = floop;

/// Mixin that causes the Widget to be listened while building. Include this
/// mixin in a StatelessWidget and override the buildGS method.
mixin _BuildGS {
  /// Override this method as you would normally override the `build` method.
  /// Do NOT override `build`.
  Widget buildGS(BuildContext context);

  /// Do NOT override this method, use [buildGS] to build your widget.
  @visibleForOverriding
  Widget build(BuildContext context) {
    fullController.startListening(context);
    var widget = buildGS(context);
    fullController.stopListening();
    return widget;
  }
}

/// Experimental lighter version of GS. It only allows reading from one
/// [Observed] while building the Widget. It's faster than BuildGS.
mixin _BuildGSLight {
  /// Override this method as you would normally override the `build` method.
  /// Do NOT override `build`.
  Widget buildGS(BuildContext context);

  /// Do NOT override this method, use [buildGS] to build your widget.
  @visibleForOverriding
  Widget build(BuildContext context) {
    lightController.startListening(context);
    var widget = buildGS(context);
    lightController.stopListening();
    return widget;
  }
}

mixin _ElementCreator on StatelessWidget {
  @override
  StatelessElement createElement() {
    return _StatelessElementGS(this);
  }
}

mixin _ElementUnmounter on StatelessElement {
  @override
  void unmount() {
    unsubscribeElement(this);
    super.unmount();
  }
}

class _StatelessElementGS = StatelessElement with _ElementUnmounter;

/// StatelessWidget class that includes GlobalState. If the Widget is built
/// reading values from an [ObservedMap] (a 'store'), the Widget will be
/// automatically rebuild when the read value on the [ObservedMap] changes.
abstract class StatelessWidgetGS = StatelessWidget
    with _BuildGS, _ElementCreator;

/// Widget that is a faster version (fewer listening time overhead) of
/// StatelessWidgetGS at the cost of allowing reads from at most one
/// [Observed] (store) while building the Widget.
///
/// This class should satisfy most uses cases, but if unsure, use the regular
/// [StatelessWidgetGS].
///
/// An example when this class should NOT be used is:
/// `store['myValue'][myNestedValue]`
/// Two [Observed] are being read, `store` and `store['myValue]`.
abstract class StatelessWidgetGSLight = StatelessWidget
    with _BuildGSLight, _ElementCreator;

/// Mixin that unsubscribes the corresponding Element when disposing the State.
mixin _StateDisposer<T extends StatefulWidget> on State<T> {
  @override
  @mustCallSuper
  dispose() {
    unsubscribeElement(this.context);
    super.dispose();
  }
}

/// State class that includes GlobalState. If the Widget is built reading
/// values from an [ObservedMap] (a 'store'), the Widget will be automatically
/// rebuild when the read value on the [ObservedMap] changes.
abstract class StateGS<T extends StatefulWidget> = State<T>
    with _BuildGS, _StateDisposer<T>;
